# Standard library
import asyncio
import contextlib
import functools
import time
from typing import (
    Callable,
    List,
    Tuple,
)

# Local libraries
from tracers.analyzers import (
    analyze_loop_snapshots,
    analyze_stack,
)
from tracers.constants import (
    LOOP_CHECK_INTERVAL,
)
from tracers.containers import (
    Frame,
    LoopSnapshot,
)
from tracers.contextvars import (
    LEVEL,
    LOOP_SNAPSHOTS,
    reset_all as reset_all_contextvars,
    STACK,
    TRACING,
)
from tracers.utils import (
    divide,
    get_function_id,
    increase_counter,
)


def measure_loop_skew(clock_id: int):

    def callback_handler(
        wanted_tick_duration: float,
        start_timestamp: float,
    ):
        timestamp: float = time.clock_gettime(clock_id)
        real_tick_duration: float = timestamp - start_timestamp
        block_duration_ratio: float = divide(
            numerator=real_tick_duration,
            denominator=wanted_tick_duration,
            on_zero_denominator=1.0,
        )

        LOOP_SNAPSHOTS.get().append(LoopSnapshot(
            block_duration_ratio=block_duration_ratio,
            wanted_tick_duration=wanted_tick_duration,
            real_tick_duration=real_tick_duration,
            timestamp=start_timestamp,
        ))

        if LEVEL.get() == 1:
            schedule_callback(wanted_tick_duration)

    def schedule_callback(wanted_tick_duration: float):
        with contextlib.suppress(RuntimeError):
            callback_handler_args: Tuple[float, float] = (
                wanted_tick_duration, time.clock_gettime(clock_id),
            )

            loop = asyncio.get_running_loop()
            loop.call_later(
                wanted_tick_duration,
                callback_handler,
                *callback_handler_args
            )

    if LEVEL.get() == 1:
        schedule_callback(LOOP_CHECK_INTERVAL)


def record_event(
    clock_id: int,
    event: str,
    function: Callable,
    function_name: str = '',
):
    _function_name, _function_signature = get_function_id(function)

    STACK.get().append(Frame(
        event=event,
        function=function_name or _function_name,
        level=LEVEL.get(),
        signature=_function_signature,
        timestamp=time.clock_gettime(clock_id),
    ))


def _get_wrapper(  # noqa: MC001
    *,
    clock_id: int = time.CLOCK_MONOTONIC,
    do_trace: bool = True,
    function_name: str = '',
    loop_snapshots_analyzer: Callable[[List[LoopSnapshot]], None] =
    analyze_loop_snapshots,
    stack_analyzer: Callable[[List[Frame]], None] =
    analyze_stack,
) -> Callable:

    def decorator(function: Callable) -> Callable:

        if not do_trace or not TRACING.get():

            # No overhead is introduced
            wrapper = function

            # Any downstream tracer is also disabled
            TRACING.set(False)

        elif asyncio.iscoroutinefunction(function):

            @functools.wraps(function)
            async def wrapper(*args, **kwargs):
                if LEVEL.get() == 0:
                    reset_all_contextvars()

                with increase_counter(LEVEL):
                    measure_loop_skew(clock_id)
                    record_event(clock_id, 'call', function, function_name)
                    result = await function(*args, **kwargs)
                    record_event(clock_id, 'return', function, function_name)

                if LEVEL.get() == 0:
                    stack_analyzer(STACK.get())
                    loop_snapshots_analyzer(LOOP_SNAPSHOTS.get())

                return result

        elif callable(function):

            @functools.wraps(function)
            def wrapper(*args, **kwargs):
                if LEVEL.get() == 0:
                    reset_all_contextvars()

                with increase_counter(LEVEL):
                    record_event(clock_id, 'call', function, function_name)
                    result = function(*args, **kwargs)
                    record_event(clock_id, 'return', function, function_name)

                if LEVEL.get() == 0:
                    stack_analyzer(STACK.get())

                return result

        else:

            # We were not able to wrap this object
            wrapper = function

        return wrapper

    return decorator


def trace_process(
    *,
    do_trace: bool = True,
    function_name: str = '',
    loop_snapshots_analyzer: Callable[[List[LoopSnapshot]], None] =
    analyze_loop_snapshots,
    stack_analyzer: Callable[[List[Frame]], None] =
    analyze_stack,
):
    return _get_wrapper(
        clock_id=time.CLOCK_PROCESS_CPUTIME_ID,
        do_trace=do_trace,
        function_name=function_name,
        loop_snapshots_analyzer=loop_snapshots_analyzer,
        stack_analyzer=stack_analyzer,
    )


def trace_thread(
    *,
    do_trace: bool = True,
    function_name: str = '',
    loop_snapshots_analyzer: Callable[[List[LoopSnapshot]], None] =
    analyze_loop_snapshots,
    stack_analyzer: Callable[[List[Frame]], None] =
    analyze_stack,
):
    return _get_wrapper(
        clock_id=time.CLOCK_THREAD_CPUTIME_ID,
        do_trace=do_trace,
        function_name=function_name,
        loop_snapshots_analyzer=loop_snapshots_analyzer,
        stack_analyzer=stack_analyzer,
    )


def trace_monotonic(
    *,
    do_trace: bool = True,
    function_name: str = '',
    loop_snapshots_analyzer: Callable[[List[LoopSnapshot]], None] =
    analyze_loop_snapshots,
    stack_analyzer: Callable[[List[Frame]], None] =
    analyze_stack,
):
    return _get_wrapper(
        clock_id=time.CLOCK_MONOTONIC,
        do_trace=do_trace,
        function_name=function_name,
        loop_snapshots_analyzer=loop_snapshots_analyzer,
        stack_analyzer=stack_analyzer,
    )


def trace(
    function: Callable = None,
    *,
    do_trace: bool = True,
    function_name: str = '',
    loop_snapshots_analyzer: Callable[[List[LoopSnapshot]], None] =
    analyze_loop_snapshots,
    stack_analyzer: Callable[[List[Frame]], None] =
    analyze_stack,
):
    wrapper = trace_monotonic(
        do_trace=do_trace,
        function_name=function_name,
        loop_snapshots_analyzer=loop_snapshots_analyzer,
        stack_analyzer=stack_analyzer,
    )

    if function:
        return wrapper(function)

    return wrapper



# Standard library
import asyncio
from collections import deque
import contextlib
from contextvars import (
    Token,
)
import inspect
from typing import (
    Callable,
    Tuple,
)

# Third party imports
from more_itertools import iter_except


def delta(start_timestamp: float, end_timestamp: float):
    return end_timestamp - start_timestamp


def divide(
    *,
    numerator: float,
    denominator: float,
    on_zero_denominator: float,
) -> float:
    return \
        on_zero_denominator if denominator == 0.0 else numerator / denominator


def drain_queue(queue: deque) -> tuple:
    return tuple(iter_except(queue.pop, IndexError))


def get_function_id(function: Callable) -> Tuple[str, str]:
    # Adding decorators to a function modify its metadata
    #   Fortunately functools' wrapped functions keep a reference to the parent
    while hasattr(function, '__wrapped__'):
        function = getattr(function, '__wrapped__')

    signature: str = '(...)'
    with contextlib.suppress(ValueError):
        signature = str(inspect.signature(function))

    module: str = function.__module__
    name: str = function.__name__
    prefix = 'async ' * asyncio.iscoroutinefunction(function)
    signature = signature if len(signature) < 48 else '(...)'

    if module not in {'__main__'}:
        return f'{prefix}{module}.{name}', signature

    return f'{prefix}{name}', signature


@contextlib.contextmanager
def increase_counter(contextvar):
    token: Token = contextvar.set(contextvar.get() + 1)
    try:
        yield
    finally:
        contextvar.reset(token)


def log(*parts):
    print(*parts)



# Standard library
import io
import contextlib
import operator
from typing import (
    List,
    NamedTuple,
)
from uuid import uuid4

# Local libraries
from tracers.containers import (
    DaemonResult,
    Frame,
    LoopSnapshot,
)
from tracers.constants import (
    CHAR_BROKEN_BAR,
    CHAR_CHECK_MARK,
    CHAR_INFO,
    CHAR_SPACE,
    CHAR_SUPERSCRIPT_ONE,
    LOOP_SKEW_TOLERANCE,
)
from tracers.utils import (
    delta,
    divide,
    log,
)
from tracers.daemon import (
    send_result_to_daemon,
)

Result = NamedTuple('Result', [
    ('counter', int),
    ('relative_timestamp', float),
    ('raw_time_ratio', float),
    ('raw_time_seconds', float),
    ('indentation', str),
    ('level', int),
    ('function', str),
])


def analyze_loop_snapshots(
    snapshots: List[LoopSnapshot],
):
    top_snapshots: List[LoopSnapshot] = sorted(
        [
            snapshot
            for snapshot in snapshots
            if snapshot.block_duration_ratio > 1.0 + LOOP_SKEW_TOLERANCE
        ],
        key=operator.attrgetter('real_tick_duration'),
        reverse=True,
    )

    if top_snapshots:
        log()
        log('  Some blocks (skews) occurred in the event loop',
            CHAR_SUPERSCRIPT_ONE)
        log()
        log('  #    Timestamp     Delay')
        log()
        initial_timestamp: float = snapshots[0].timestamp
        for counter, snapshot in enumerate(top_snapshots):
            skew: float = delta(
                snapshot.wanted_tick_duration,
                snapshot.real_tick_duration,
            )

            log(f'{counter:>6}',
                f'{delta(initial_timestamp, snapshot.timestamp):>8.2f}s',
                f'{skew:>8.2f}s')
        log()
        log(f'  {CHAR_SUPERSCRIPT_ONE}',
            'Consider reviewing them carefully',
            'to improve the overall system throughput')


def analyze_stack(stack: List[Frame]):
    with io.StringIO() as buffer:
        with contextlib.redirect_stdout(buffer):
            _analyze_stack(stack)

        buffer.seek(0)
        stdout = buffer.read()

    log(stdout)
    send_result_to_daemon(DaemonResult(
        stack=stack,
        stdout=stdout,
    ))


def _analyze_stack(stack: List[Frame]):
    stack_levels: List[int] = \
        list(map(operator.attrgetter('level'), stack))

    total_time_seconds: float = \
        delta(stack[0].timestamp, stack[-1].timestamp)

    log()
    log(f'{CHAR_INFO} Finished transaction: {uuid4().hex},',
        f'{total_time_seconds:.2f} seconds')
    log()
    log('  #    Timestamp       %     Total    Nested Call Chain')
    log()

    counter: int = 0
    results: List[Result] = []
    accumulator: List[Result] = []
    for index, frame in enumerate(stack):
        if frame.event == 'call':
            counter += 1

            frame_childs: List[Frame] = \
                stack[index:stack_levels.index(frame.level, index + 1) + 1]

            raw_time_seconds: float = \
                delta(frame_childs[0].timestamp, frame_childs[-1].timestamp)

            results.append(Result(
                counter=counter,
                relative_timestamp=delta(stack[0].timestamp, frame.timestamp),
                raw_time_ratio=100.0 * divide(
                    numerator=raw_time_seconds,
                    denominator=total_time_seconds,
                    on_zero_denominator=1.0,
                ),
                raw_time_seconds=raw_time_seconds,
                level=frame.level,
                indentation=(
                    (CHAR_SPACE * 3 + CHAR_BROKEN_BAR) * (frame.level - 1) +
                    (CHAR_SPACE * 3 + CHAR_CHECK_MARK)
                ),
                function=frame.function,
            ))

    if results:
        accumulator.append(results[0])
        flush_accumulator(accumulator)

    for index, result in enumerate(results[1:-1], start=1):
        accumulator.append(result)

        if ((result.level != results[index - 1].level or
                result.function != results[index - 1].function) and
            (result.level != results[index + 1].level or
                result.function != results[index + 1].function)):
            flush_accumulator(accumulator)

    if len(results) > 1:
        accumulator.append(results[-1])
        flush_accumulator(accumulator)


def flush_accumulator(accumulator: List[Result]):
    if accumulator:
        times: str = f'{len(accumulator)} times: ' * (len(accumulator) > 1)

        log(
            f'{accumulator[0].counter:>6}',
            f'{accumulator[0].relative_timestamp:>8.2f}s',
            f'{sum(r.raw_time_ratio for r in accumulator):>6.1f}%',
            f'{sum(r.raw_time_seconds for r in accumulator):>8.2f}s',
            f'{accumulator[0].indentation}',
            f'{times}{accumulator[0].function}',
        )
        accumulator.clear()



# Standard library
import asyncio
from collections import deque
from threading import Thread

# Local libraries
from tracers.utils import (
    drain_queue,
)

# Private constants
_RESULTS_QUEUE: deque = deque()


async def daemon():
    while True:
        await asyncio.sleep(1.0)
        len(drain_queue(_RESULTS_QUEUE))


def send_result_to_daemon(result):
    _RESULTS_QUEUE.appendleft(result)


# Side effect: Start an asynchronous daemon server
Thread(daemon=True, target=lambda: asyncio.run(daemon())).start()



# Standard library
from contextvars import (
    ContextVar,
)
from typing import (
    Any,
    Callable,
    List,
    Tuple,
)

# Local libraries
from tracers.containers import (
    Frame,
    LoopSnapshot,
)

LEVEL: ContextVar[int] = \
    ContextVar('LEVEL', default=0)
LOOP_SNAPSHOTS: ContextVar[List[LoopSnapshot]] = \
    ContextVar('LOOP_SNAPSHOTS')
STACK: ContextVar[List[Frame]] = \
    ContextVar('STACK')
TRACING: ContextVar[bool] = \
    ContextVar('TRACING', default=True)

RESETABLE: Tuple[Tuple[ContextVar, Callable[[], Any]], ...] = (
    (LEVEL, lambda: 0),
    (LOOP_SNAPSHOTS, lambda: []),
    (STACK, lambda: []),
)


def reset_all():
    for var, default_value_creator in RESETABLE:
        reset_one(var, default_value_creator)


def reset_one(
    contextvar: ContextVar,
    default_value_creator: Callable[[], Any],
) -> None:
    obj = contextvar.get(None)

    if obj:
        if isinstance(obj, list):
            # Delete all elements
            obj.clear()
        elif isinstance(obj, (bool, int)):
            # This never leaks
            pass
        else:
            raise TypeError(f'Missing handler for type: {type(obj)}')

    # Delete reference
    del obj

    # Re-initialize to its default value
    contextvar.set(default_value_creator())
